/*
	MegaCracker - Cracks the password hash from a confirmation link.
	Copyright (C) 2013 Steve Thomas <steve AT tobtu DOT com>

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include <stdlib.h>
#include <string.h>
#include "megahashgenni.h"
#include "aes128.h"

MegaHashGen_NI::MegaHashGen_NI(
	MessageQueue<msgQueueNodeType> *inQueue,
	MessageQueue<msgQueueNodeType> *outQueue)
	: m_inQueue(inQueue),
	  m_outQueue(outQueue),
	  m_run(1)
{
	if (inQueue == NULL || outQueue == NULL)
	{
		reportError("MegaHashGen_NI::MegaHashGen_NI(): NULL pointer");
		exit(1);
	}
}

MegaHashGen_NI::~MegaHashGen_NI()
{
	if (!stopRunning())
	{
		waitForStopRunning();
	}
}

int MegaHashGen_NI::stopRunning()
{
	if (m_run)
	{
		m_run = 0;
		m_inQueue->flush();
		return 0;
	}
	return 1;
}

void *MegaHashGen_NI::threadFunc()
{
#ifdef ARC_x86

#define SWAP_ENDIAN(n) _mm_shuffle_epi8(n, _mm_set_epi32(0x0c0d0e0f,0x08090a0b,0x04050607,0x00010203))

	__m128i ctx[AES128_CTX_SIZE * MAX_PW_BLOCKS];
	__m128i pw[PW_BUFFER_SIZE / sizeof(__m128i) + 1];
	__m128i pwKey;
	MessageQueue<msgQueueNodeType>     *inQueue  = m_inQueue;
	MessageQueue<msgQueueNodeType>     *outQueue = m_outQueue;
	messageQueueNode<msgQueueNodeType> *node;
	uint32_t *pwKeys;
	uint8_t  *passwords;

	// Get work unit
	node = inQueue->pop();
	while (node != NULL && m_run)
	{
		pwKeys    = node->data.pwKeys;
		passwords = (uint8_t*) node->data.passwords;
		for (uint32_t i = 0, count = node->data.count; i < count; i++)
		{
			// Get password
			uint32_t pwLen = *((uint16_t*) passwords);
			uint32_t pwBlocks = (pwLen + 15) / 16;

			passwords += 2;
			pw[pwLen / 16] = _mm_set1_epi32(0);
			for (uint32_t j = 0; j < pwLen; j += 4)
			{
				uint32_t tmp = passwords[j] << 24;
				if (j + 1 < pwLen)
				{
					tmp |= passwords[j + 1] << 16;
					if (j + 2 < pwLen)
					{
						tmp |= passwords[j + 2] << 8;
						if (j + 3 < pwLen)
						{
							tmp |= passwords[j + 3];
						}
					}
				}
				((uint32_t*) pw)[j >> 2] = tmp;
			}
			passwords += pwLen;

			// MegaKDF
			pwKey = _mm_set_epi32(0x56cb5201, 0x813fbed1, 0xa4c7b07d, 0xe367c493);
			if (pwBlocks)
			{
				newAes128Encrypt_NI(pw, ctx);
				for (uint32_t j = 1; j < pwBlocks; j++)
				{
					newAes128Encrypt_NI(pw + j, ctx + AES128_CTX_SIZE / 4 * j);
				}
				for (uint32_t j = 0; j < 65536; j++)
				{
					encryptAes128_NI(ctx, &pwKey);
					for (uint32_t k = 1; k < pwBlocks; k++)
					{
						encryptAes128_NI(ctx + AES128_CTX_SIZE / 4 * k, &pwKey);
					}
				}
			}
			pwKey = SWAP_ENDIAN(pwKey);
			memcpy(pwKeys, &pwKey, sizeof(pwKey));
			pwKeys += sizeof(pwKey) / sizeof(uint32_t);
		}

		// Get next work unit
		outQueue->push(node);
		node = inQueue->pop();
	}
#else
	reportError("MegaHashGen_NI::threadFunc() You are trying to run x86 code on a non-x86 CPU.");
	exit(1);
#endif
	m_run = 0;
	return NULL;
}
