/*
	MegaCracker - Cracks the password hash from a confirmation link.
	Copyright (C) 2013 Steve Thomas <steve AT tobtu DOT com>

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#ifndef MESSAGE_QUEUE_H
#define MESSAGE_QUEUE_H

#include "common.h"
#include "threads.h"

template <class T>
struct messageQueueNode
{
	messageQueueNode<T> *next;
	T                    data;
};

template <class T>
class MessageQueue
{
public:
	MessageQueue();
	virtual ~MessageQueue();
	virtual messageQueueNode<T> *pop();
	messageQueueNode<T> *popAll();
	void push(messageQueueNode<T> *node);
	void pushQueue(messageQueueNode<T> *head, messageQueueNode<T> *tail);
	void flush();
	void stopFlush();
	void waitForStopFlush();

protected:
	MessageQueue(MessageQueue<T> &q) {}
	MessageQueue<T> &operator=(MessageQueue<T> &q) {return *this;}

	messageQueueNode<T> *p_head, *p_tail;
	MUTEX    p_mutex;
	COND     p_cond, p_cond2;
	uint32_t p_cleanUp, p_flushing, p_numWaiting;
};

template <class T>
MessageQueue<T>::MessageQueue()
{
	p_head = NULL;
	p_tail = NULL;
	p_flushing = 0;
	p_numWaiting = 0;
	MUTEX_CREATE(p_mutex);
	COND_CREATE(p_cond);
	COND_CREATE(p_cond2);
}

template <class T>
MessageQueue<T>::~MessageQueue()
{
	MUTEX_DELETE(p_mutex);
	COND_DELETE(p_cond);
	COND_DELETE(p_cond2);
}

template <class T>
messageQueueNode<T> *MessageQueue<T>::pop()
{
	messageQueueNode<T> *ret = NULL;

	MUTEX_LOCK(p_mutex);
	while (p_head == NULL && p_flushing == 0)
	{
		COND_WAIT(p_cond, p_mutex);
	}
	if (p_head != NULL)
	{
		ret = p_head;
		p_head = p_head->next;
		ret->next = NULL;
	}
	MUTEX_UNLOCK(p_mutex);
	return ret;
}

template <class T>
messageQueueNode<T> *MessageQueue<T>::popAll()
{
	messageQueueNode<T> *ret;

	MUTEX_LOCK(p_mutex);
	while (p_head == NULL && p_flushing == 0)
	{
		COND_WAIT(p_cond, p_mutex);
	}
	ret = p_head;
	p_head = NULL;
	p_tail = NULL;
	MUTEX_UNLOCK(p_mutex);
	return ret;
}

template <class T>
void MessageQueue<T>::push(messageQueueNode<T> *n)
{
	MUTEX_LOCK(p_mutex);
	n->next = NULL;
	if (p_head == NULL)
	{
		COND_SIGNAL(p_cond);
		p_head = n;
	}
	else
	{
		p_tail->next = n;
	}
	p_tail = n;
	MUTEX_UNLOCK(p_mutex);
}

template <class T>
void MessageQueue<T>::pushQueue(messageQueueNode<T> *h, messageQueueNode<T> *t)
{
	MUTEX_LOCK(p_mutex);
	if (p_head == NULL)
	{
		COND_SIGNAL_ALL(p_cond);
		p_head = h;
	}
	else
	{
		p_tail->next = h;
	}
	p_tail = t;
	p_tail->next = NULL;
	MUTEX_UNLOCK(p_mutex);
}

template <class T>
void MessageQueue<T>::flush()
{
	MUTEX_LOCK(p_mutex);
	p_flushing = 1;
	COND_SIGNAL_ALL(p_cond);
	MUTEX_UNLOCK(p_mutex);
}

template <class T>
void MessageQueue<T>::stopFlush()
{
	MUTEX_LOCK(p_mutex);
	p_flushing = 0;
	COND_SIGNAL_ALL(p_cond2);
	MUTEX_UNLOCK(p_mutex);
}

template <class T>
void MessageQueue<T>::waitForStopFlush()
{
	MUTEX_LOCK(p_mutex);
	p_numWaiting++;
	while (p_flushing != 0)
	{
		COND_WAIT(p_cond2, p_mutex);
	}
	p_numWaiting--;
	MUTEX_UNLOCK(p_mutex);
}

#endif
